@@ -7,8 +7,9 @@ use Carp;
 use IO::File;
 use IO::Socket;
 use IO::Select;
+use Net::IMAP::Simple::PipeSocket;
 
-our $VERSION = "1.2010_99";
+our $VERSION = "1.2034_1";
 
 BEGIN {
     # I'd really rather the pause/cpan indexers miss this "package"
@@ -18,6 +19,8 @@ BEGIN {
        sub new { bless $_[1] })
 }
 
+our $uidm;
+
 sub new {
     my ( $class, $server, %opts ) = @_;
 
@@ -34,6 +37,11 @@ sub new {
             if $opts{use_v6} and $opts{use_ssl};
     }
 
+    if( $opts{ssl_version} ) {
+        $self->{ssl_version} = $opts{ssl_version};
+        $opts{use_ssl} = 1;
+    }
+
     if( $opts{use_ssl} ) {
         eval {
             require IO::Socket::SSL;
@@ -52,25 +60,34 @@ sub new {
         } or croak "IO::Socket::INET6 must be installed in order to use_v6";
     }
 
-    my ( $srv, $prt ) = split( /:/, $server, 2 );
-    $prt ||= ( $opts{port} ? $opts{port} : $self->_port );
+    if( $server =~ m/cmd:(.+)/ ) {
+        $self->{cmd} = $1;
 
-    $self->{server} = $srv;
-    $self->{port}   = $prt;
+    } else {
+        if( ($self->{server}, $self->{port}) = $server =~ m/^(\d{1,3}(?:\.\d{1,3}){3})(?::(\d+))?\z/ ) {
 
-    $self->{timeout}          = ( $opts{timeout} ? $opts{timeout} : $self->_timeout );
-    $self->{retry}            = ( $opts{retry} ? $opts{retry} : $self->_retry );
-    $self->{retry_delay}      = ( $opts{retry_delay} ? $opts{retry_delay} : $self->_retry_delay );
-    $self->{bindaddr}         = $opts{bindaddr};
-    $self->{use_select_cache} = $opts{use_select_cache};
-    $self->{select_cache_ttl} = $opts{select_cache_ttl};
-    $self->{debug}            = $opts{debug};
+        } elsif( ($self->{server}, $self->{port}) = $server =~ m/^\[([a-fA-F0-9:]+)\]:(\d+)\z/ ) {
 
-    # Pop the port off the address string if it's not an IPv6 IP address
-    if ( !$self->{use_v6} && $self->{server} =~ /^[A-Fa-f0-9]{4}:[A-Fa-f0-9]{4}:/ && $self->{server} =~ s/:(\d+)$//g ) {
-        $self->{port} = $1;
+        } elsif( ($self->{server}, $self->{port}) = $server =~ m/^([a-fA-F0-9:]+)\z/ ) {
+
+        } elsif( ($self->{server}, $self->{port}) = $server =~ m/^([^:]+):(\d+)\z/ ) {
+
+        } else {
+            $self->{server} = $server;
+        }
+
+        $self->{port} = $self->_port unless defined $self->{port};
     }
 
+    $self->{timeout}           = ( $opts{timeout} ? $opts{timeout} : $self->_timeout );
+    $self->{retry}             = ( defined($opts{retry}) ? $opts{retry} : $self->_retry );
+    $self->{retry_delay}       = ( defined($opts{retry_delay}) ? $opts{retry_delay} : $self->_retry_delay );
+    $self->{bindaddr}          = $opts{bindaddr};
+    $self->{use_select_cache}  = $opts{use_select_cache};
+    $self->{select_cache_ttl}  = $opts{select_cache_ttl};
+    $self->{debug}             = $opts{debug};
+    $self->{readline_callback} = $opts{readline_callback};
+
     my $sock;
     my $c;
     for ( my $i = 0 ; $i <= $self->{retry} ; $i++ ) {
@@ -97,10 +114,11 @@ sub new {
 
     my $select = $self->{sel} = IO::Select->new($sock);
 
-    $self->_debug( caller, __LINE__, 'new', "looking for greeting" ) if $self->{debug};
+    $self->_debug( caller, __LINE__, 'new', "waiting for socket ready" ) if $self->{debug};
 
     my $greeting_ok = 0;
     if( $select->can_read($self->{timeout}) ) {
+        $self->_debug( caller, __LINE__, 'new', "looking for greeting" ) if $self->{debug};
         if( my $line = $sock->getline ) {
             # Cool, we got a line, check to see if it's a
             # greeting.
@@ -132,21 +150,30 @@ sub _connect {
     my ($self) = @_;
     my $sock;
 
-    $sock = $self->_sock_from->new(
-        PeerAddr => $self->{server},
-        PeerPort => $self->{port},
-        Timeout  => $self->{timeout},
-        Proto    => 'tcp',
-        ( $self->{bindaddr} ? { LocalAddr => $self->{bindaddr} } : () )
-    );
+    if( $self->{cmd} ) {
+        $self->_debug( caller, __LINE__, '_connect', "popping open a pipesocket for command: $self->{cmd}" ) if $self->{debug};
+        $sock = Net::IMAP::Simple::PipeSocket->new(cmd=>$self->{cmd});
 
+    } else {
+        $self->_debug( caller, __LINE__, '_connect', "connecting to $self->{server}:$self->{port}" ) if $self->{debug};
+        $sock = $self->_sock_from->new(
+            PeerAddr => $self->{server},
+            PeerPort => $self->{port},
+            Timeout  => $self->{timeout},
+            Proto    => 'tcp',
+            ( $self->{bindaddr} ? ( LocalAddr => $self->{bindaddr} ) : () ),
+            ( $_[0]->{ssl_version} ? (SSL_version => $self->{ssl_version}) : ()),
+        );
+    }
+
+    $self->_debug( caller, __LINE__, '_connect', "connected, returning socket" ) if $self->{debug};
     return $sock;
 }
 
-sub _port        { return $_[0]->{use_ssl} ? 993 : 143 } 
+sub _port        { return $_[0]->{use_ssl} ? 993 : 143 }
 sub _sock        { return $_[0]->{sock} }
 sub _count       { return $_[0]->{count} }
-sub _last        { $_[0]->select unless exists $_[0]->{last}; return $_[0]->{last} }
+sub _last        { $_[0]->select unless exists $_[0]->{last}; return $_[0]->{last}||0 }
 sub _timeout     { return 90 }
 sub _retry       { return 1 }
 sub _retry_delay { return 5 }
@@ -170,7 +197,7 @@ sub starttls {
 
             my $startres = IO::Socket::SSL->start_SSL(
                 $self->{sock},
-                SSL_version        => "SSLv3 TLSv1",
+                SSL_version        => $self->{ssl_version} || "SSLv3 TLSv1",
                 SSL_startHandshake => 0,
             );
 
@@ -189,13 +216,27 @@ sub starttls {
 sub login {
     my ( $self, $user, $pass ) = @_;
 
+    $pass = _escape($pass);
+
     return $self->_process_cmd(
-        cmd     => [ LOGIN => qq[$user "$pass"] ],
+        cmd     => [ LOGIN => qq[$user $pass] ],
         final   => sub { 1 },
         process => sub { },
     );
 }
 
+sub separator {
+    my ( $self, ) = @_;
+    my $sep;
+
+        return $self->_process_cmd (
+        cmd     => [ LIST => qq["" ""]  ],
+        final => sub { $sep },
+        process => sub { (undef,undef,undef,$sep,undef) = split /\s/smx , $_[0];
+                        $sep =~ s/["]//g;  },
+    );
+}
+
 sub _clear_cache {
     my $self = shift;
 
@@ -225,26 +266,28 @@ sub uidvalidity {
     return $self->status($mbox => 'uidvalidity');
 }
 
+sub uidsearch {
+    my $self = shift;
+
+    local $uidm = 1;
+
+    return $self->search(@_);
+}
+
 sub uid {
+    my $self = shift;
+       $self->_be_on_a_box; # does a select if we're not on a mailbox
+
+    return $self->uidsearch( shift || "1:*" );
+}
+
+sub seq {
     my $self = shift;
     my $msgno = shift || "1:*";
 
     $self->_be_on_a_box; # does a select if we're not on a mailbox
 
-    my @_uids;
-
-    # 3 UID SEARCH 28,4,30\r\n
-    # * SEARCH 58864 58888 58890\r\n
-
-    return $self->_process_cmd(
-        cmd     => [ "UID SEARCH " . _escape($msgno) ],
-        final   => sub { wantarray ? @_uids : $_uids[-1] },
-        process => sub {
-            if( my ($digits) = $_[0] =~ m/\* SEARCH\s+([\d\s]+)/i ) {
-                @_uids = split m/\s+/, $digits;
-            }
-        },
-    );
+    return $self->search("uid $msgno");
 }
 
 sub status {
@@ -443,12 +486,44 @@ sub deleted {
     return 0;
 }
 
+sub range2list {
+    my $self_or_class = shift;
+    my %h;
+    my @items = grep {!$h{$_}++} map { m/(\d+):(\d+)/ ? ($1 .. $2) : ($_) } split(m/[,\s]+/, shift);
+
+    return @items;
+}
+
+sub list2range {
+    my $self_or_class = shift;
+    my %h;
+    my @a = sort { $a<=>$b } grep {!$h{$_}++} grep {m/^\d+/} grep {defined $_} @_;
+    my @b;
+
+    while(@a) {
+        my $e = 0;
+
+        $e++ while $e+1 < @a and $a[$e]+1 == $a[$e+1];
+
+        push @b, ($e>0 ? [$a[0], $a[$e]] : [$a[0]]);
+        splice @a, 0, $e+1;
+    }
+
+    return join(",", map {join(":", @$_)} @b);
+}
+
 sub list {
     my ( $self, $number ) = @_;
 
+    # NOTE: this entire function is horrible:
+    # 1. it should be called message_size() or something similar
+    # 2. what if $number is a range? none of this works right
+
     my $messages = $number || '1:' . $self->_last;
     my %list;
 
+    return {} if $messages eq "1:0";
+
     return $self->_process_cmd(
         cmd => [ FETCH => qq[$messages RFC822.SIZE] ],
         final => sub { $number ? $list{$number} : \%list },
@@ -462,16 +537,18 @@ sub list {
 
 sub search {
     my ($self, $search, $sort, $charset) = @_;
+
     $search   ||= "ALL";
     $charset  ||= 'UTF-8';
-    my $cmd   = 'SEARCH';
+
+    my $cmd = $uidm ? 'UID SEARCH' : 'SEARCH';
 
     $self->_be_on_a_box; # does a select if we're not on a mailbox
 
     # add rfc5256 sort, requires charset :(
     if ($sort) {
         $sort = uc $sort;
-        $cmd = "SORT ($sort) \"$charset\"";
+        $cmd = ($uidm ? "UID ": "") . "SORT ($sort) \"$charset\"";
     }
 
     my @seq;
@@ -480,8 +557,10 @@ sub search {
         cmd => [ $cmd => $search ],
         final => sub { wantarray ? @seq : int @seq },
         process => sub { if ( my ($msgs) = $_[0] =~ /^\*\s+(?:SEARCH|SORT)\s+(.*)/i ) {
-            push @seq, $1 while $msgs =~ m/\b(\d+)\b/g;
-        } },
+
+            @seq = $self->range2list($msgs);
+
+        }},
     );
 }
 
@@ -548,12 +627,52 @@ sub get {
     my $arg = $part ? "BODY[$part]" : 'RFC822';
 
     my @lines;
+    my $fetching;
 
     return $self->_process_cmd(
         cmd => [ FETCH => qq[$number $arg] ],
-        final => sub { pop @lines; wantarray ? @lines : Net::IMAP::Simple::_message->new(\@lines) },
+        final => sub {
+            if( $fetching ) {
+                if( $fetching > 0 ) {
+                    # XXX: this is just about the least efficient way in the
+                    # world to do this; I should appologize, but really,
+                    # nothing in this module is done particularly well.  I
+                    # doubt anyone will notice this.
+
+                    local $"="";
+                    my $message = "@lines";
+                    @lines = split m/(?<=\x0d\x0a)/, substr($message, 0, $fetching)
+                        if( length $message > $fetching );
+                }
+                return  wantarray ? @lines : Net::IMAP::Simple::_message->new(\@lines)
+            }
+
+            if( defined $fetching and $fetching == 0 ) {
+                return "\n"; # XXX: Your 0 byte message is incorrectly returned as a newline.  Meh.
+            }
+
+            # NOTE: There is not supposed to be an error if you ask for a
+            # message that's not there, but this is a rather confusing
+            # notion … so we generate an error here.
+
+            $self->{_errstr} = "message not found";
+            return;
+        },
         process => sub {
-            if ( $_[0] !~ /^\* \d+ FETCH/ ) {
+            if ( $_[0] =~ /^\*\s+\d+\s+FETCH\s+\(.+?\{(\d+)\}/ ) {
+                $fetching = $1;
+
+            } elsif( $_[0] =~ /^\*\s+\d+\s+FETCH\s+\(.+?\"(.*)\"\s*\)/ ) {
+                # XXX: this is not tested because Net::IMAP::Server doesn't do
+                # this type of string result (that I know of) for this it might
+                # work, ... frog knows.  Not likely to come up very often, if
+                # ever; although you do sometimes see the occasional 0byte
+                # message.  Valid really.
+
+                $fetching = -1;
+                @lines = ($1);
+
+            } elsif( $fetching ) {
                 push @lines, join( ' ', @_ );
             }
         },
@@ -563,10 +682,9 @@ sub get {
 
 sub _process_flags {
     my $self = shift;
+    my @ret = map { split m/\s+/, $_ } grep { $_ } @_;
 
-    return grep { m/^\\\w+\z/ }
-            map { split m/\s+/, $_ }
-            @_;
+    return @ret;
 }
 
 sub put {
@@ -586,16 +704,18 @@ sub put {
         cmd   => [ APPEND => _escape($mailbox_name) ." (@flags) {$size}" ],
         final => sub { $self->_clear_cache },
         process => sub {
-            if ($size) {
-                my $sock = $self->_sock;
-                if ( ref $msg eq "ARRAY" ) {
-                    print $sock $_ for @$msg;
+            if( $_[0] =~ m/^\+\s+/ ) { # + continue (or go ahead, or whatever)
+                if ($size) {
+                    my $sock = $self->_sock;
+                    if ( ref $msg eq "ARRAY" ) {
+                        print $sock $_ for @$msg;
 
-                } else {
-                    print $sock $msg;
+                    } else {
+                        print $sock $msg;
+                    }
+                    $size = undef;
+                    print $sock "\r\n";
                 }
-                $size = undef;
-                print $sock "\r\n";
             }
         },
 
@@ -908,6 +1028,17 @@ sub copy {
     );
 }
 
+sub uidcopy {
+    my ( $self, $number, $box ) = @_;
+    my $b = _escape($box);
+
+    return $self->_process_cmd(
+        cmd     => [ 'UID COPY' => qq[$number $b] ],
+        final   => sub { 1 },
+        process => sub { },
+    );
+}
+
 sub waserr {
     return $_[0]->{_waserr};
 }
@@ -979,8 +1110,14 @@ sub _read_multiline {
     my $read_so_far = 0;
 
     while ( $read_so_far < $count ) {
-        push @lines, $sock->getline;
-        $read_so_far += length( $lines[-1] );
+        if( defined( my $line = $sock->getline ) ) {
+            $read_so_far += length( $line );
+            push @lines, $line;
+
+        } else {
+            $self->_seterrstr( "error reading $count bytes from socket" );
+            last;
+        }
     }
 
     if ( $self->{debug} ) {
@@ -999,13 +1136,19 @@ sub _process_cmd {
     $args{process} = sub {} unless ref($args{process}) eq "CODE";
     $args{final}   = sub {} unless ref($args{final})   eq "CODE";
 
+    my $cb = $self->{readline_callback};
+
     my $res;
     while ( $res = $sock->getline ) {
+        $cb->($res) if $cb;
         $self->_debug( caller, __LINE__, '_process_cmd', $res ) if $self->{debug};
 
         if ( $res =~ /^\*.*\{(\d+)\}[\r\n]*$/ ) {
             $args{process}->($res);
-            $args{process}->($_) foreach $self->_read_multiline( $sock, $1 );
+            foreach( $self->_read_multiline( $sock, $1 ) ) {
+                $cb->($_) if $cb;
+                $args{process}->($_)
+            }
 
         } else {
             my $ok = $self->_cmd_ok($res);
@@ -1066,4 +1209,4 @@ sub _debug {
     return;
 }
 
-"True"; ## no critic -- pfft
+1;
